// CC BY
// code by SASANO Takayoshi

#include "commanddefs.h"
#include "serial.h"
#include "gfx.h"
#include "sprite.h"
#include "video.h"
#include "misc.h"
#include "image.h"

#define	CMDBUF_SIZE	256
#define	DECODEBUF_SIZE	64

static unsigned char	CmdBuf[CMDBUF_SIZE];
static unsigned char	DecodeBuf[DECODEBUF_SIZE];

static	void	do_clear(void *arg, unsigned char length);
static	void	do_line(void *arg, unsigned char length);
static	void	do_circle(void *arg, unsigned char length);
static	void	do_stamp(void *arg, unsigned char length);
static	void	do_stamp1(void *arg, unsigned char length);
static	void	do_image(void *arg, unsigned char length);
static	void	do_video(void *arg, unsigned char length);
static	void	do_sprite_start(void *arg, unsigned char length);
static	void	do_sprite_create(void *arg, unsigned char length);
static	void	do_sprite_move(void *arg, unsigned char length);

struct command_table {
	unsigned char	command;
	char		*string;
	void	(*do_command)(void *buf, unsigned char length);
};

static	const struct command_table CmdTable[] = {
	{COMMAND_CLEAR, STRING_CLEAR, do_clear},
	{COMMAND_LINE, STRING_LINE, do_line},
	{COMMAND_STAMP, STRING_STAMP, do_stamp},
	{COMMAND_STAMP1, STRING_STAMP1, do_stamp1},
	{COMMAND_IMAGE, STRING_IMAGE, do_image},
	{COMMAND_VIDEO, STRING_VIDEO, do_video},
	{COMMAND_SPRITE_START, STRING_SPRITE_START, do_sprite_start},
	{COMMAND_SPRITE_CREATE, STRING_SPRITE_CREATE, do_sprite_create},
	{COMMAND_SPRITE_MOVE, STRING_SPRITE_MOVE, do_sprite_move},
	{COMMAND_SOUND, STRING_SOUND, NULL},
	{COMMAND_SOUND1, STRING_SOUND1, NULL},
	{COMMAND_MUSIC_SCORE, STRING_MUSIC_SCORE, NULL},
	{COMMAND_MUSIC_PLAY, STRING_MUSIC_PLAY, NULL},
	{COMMAND_RESET, STRING_RESET, NULL},
	{COMMAND_CIRCLE, STRING_CIRCLE, do_circle},
};

#define	CmdTable_entry	(sizeof(CmdTable) / sizeof(CmdTable[0]))

#if 0
void	hexdump(unsigned char *buf, int length)
{
	int	i;
	static	char	hex[] = "0123456789ABCDEF";

	sio_send_char('\n');
	sio_send_char('\r');
	for (i = 0; i < length; i++) {
		sio_send_char(hex[(*buf >> 4) & 0x0f]);
		sio_send_char(hex[(*buf >> 0) & 0x0f]);
		sio_send_char(' ');
		buf++;
	}
	sio_send_char('\n');
	sio_send_char('\r');
}
#endif

static	void	do_clear(void *arg, unsigned char length)
{
	struct command_clear_arg	*p = arg;

	if (sprite_is_enable()) {
		return;
	}

	gfx_clear(p->cn);
	video_update();
}

static	void	do_line(void *arg, unsigned char length)
{
	struct command_line_arg	*p = arg;

	if (sprite_is_enable()) {
		return;
	}

	gfx_draw_line(p->x1, p->y1, p->x2, p->y2, p->cn);
	video_update();
}

static	void	do_circle(void *arg, unsigned char length)
{
	struct command_circle_arg *p = arg;

	if (sprite_is_enable()) {
		return;
	}

	gfx_draw_circle(p->xc, p->yc, p->ra, p->cn);
	video_update();
}

static	void	do_stamp(void *arg, unsigned char length)
{
	struct command_stamp_arg *p = arg;

	if (sprite_is_enable()) {
		return;
	}

	gfx_draw_block8x8(p->px, p->py, p->tc, p->cc);
	video_update();
}

static	void	do_stamp1(void *arg, unsigned char length)
{
	struct command_stamp1_arg *p = arg;

	if (sprite_is_enable()) {
		return;
	}

	gfx_draw_bitmap8x8(p->px, p->py, p->cn, p->cc);
	video_update();
}

static	void	do_image(void *arg, unsigned char length)
{
	struct command_image_arg *p = arg;

	if (sprite_is_enable()) {
		return;
	}

	if (p->in < image_entry) {
		gfx_display_image(get_image_data(p->in));
		video_update();
	}
}

static	void	do_video(void *arg, unsigned char length)
{
	struct command_video_arg *p = arg;

	video_enable(p->os);
}

static	void	do_sprite_start(void *arg, unsigned char length)
{
	struct	command_sprite_start_arg	*p = arg;

	sprite_set_bg(p->in);
	sprite_render();
	video_update();
}

static	void	do_sprite_create(void *arg, unsigned char length)
{
	struct	command_sprite_create_arg	*p = arg;

	sprite_define(p->sn, p->si);
	sprite_render();
	video_update();
}

static	void	do_sprite_move(void *arg, unsigned char length)
{
	struct	command_sprite_move_arg		*p = arg;

	sprite_move(p->sn, p->px, p->py);
	sprite_render();
	video_update();
}

static	unsigned char	decode_nibble(unsigned char c)
{
	if (c >= '0' && c <= '9') return c - '0';
	else if (c >= 'A' && c <= 'F') return c - 'A' + 10;
	else if (c >= 'a' && c <= 'f') return c - 'a' + 10;
	else return 0;
}

static	unsigned short decode_hex(unsigned char *str, unsigned char *buf, unsigned short bufsize)
{
	unsigned short	i, len;
	unsigned char	n;

	len = 0;
	while (*str != '\0') {
		if (!isxdigit(*str)) {
			str++;
			continue;
		}

		/* 1st character */
		n = decode_nibble(*str);
		str++;

		/* 2nd character */
		if (*str != '\0' && isxdigit(*str)) {
			n <<= 4;
			n |= decode_nibble(*str);
			str++;
		}

		*buf = n;
		buf++;

		len++;
		if (len >= bufsize) break;
	}

	for (i = len; i < bufsize; i++) {
		*buf = 0;
		buf++;
	}

	return len;
}

static	short	execute_text_command(void)
{
	unsigned short	i, n;
	void	(*func)(void *buf, unsigned char length);

	func = NULL;
	for (i = 0; i < CmdTable_entry; i++) {
		n = strlen(CmdTable[i].string);
		if (!memcmp(CmdBuf, CmdTable[i].string, n)) {
			if (CmdTable[i].command == COMMAND_RESET) {
				return 1;
			}

			func = CmdTable[i].do_command;
			n = decode_hex(&CmdBuf[n], DecodeBuf, DECODEBUF_SIZE);
			break;
		}
	}

	if (func != NULL) {
		(*func)(DecodeBuf, n);
	}

	return 0;
}

static	short	execute_binary_command(unsigned char length)
{
	unsigned short	i;
	void	(*func)(void *buf, unsigned char length);

	if (*CmdBuf == COMMAND_RESET) {
		return 1;
	}

	func = NULL;
	for (i = 0; i < CmdTable_entry; i++) {
		if (*CmdBuf == CmdTable[i].command) {
			func = CmdTable[i].do_command;
			break;
		}
	}

	if (func != NULL) {
		(*func)(CmdBuf, length);
	}

	return 0;
}

static	void	read_binary_command(unsigned char length)
{
	unsigned short	i;
	unsigned char	*buf = CmdBuf;

	for (i = 0; i < length; i++) {
		*buf = sio_recv_char();
		buf++;
	}

	for (; i < CMDBUF_SIZE; i++) {
		*buf = 0;
		buf++;
	}
}

static	void	read_text_command(void)
{
	unsigned short	i;
	unsigned char	c;

	for (i = 0; i < CMDBUF_SIZE - 1; i++) {
		c = sio_recv_char();

		/* TAB treat as space */
		if (c == 0x09) c = ' ';

		/* control character: quit */
		if (c < ' ') break;

		CmdBuf[i] = c;
	}

	/* terminate */
	CmdBuf[i] = '\0';
}

void	command_loop(void)
{
	unsigned char	c, n;
	unsigned const char	*p;
	static	const unsigned char pancake_header[] = "PANCAKE ";

	/* startup screen */
	sprite_init();
	gfx_display_image(get_image_data(0));
	video_update();

	p = pancake_header;

	while (1) {
		c = sio_recv_char();

		if (c == 0x80) {		// binary command
			n = sio_recv_char();	// length (data + header)
			if (n > 2) {
				n -= 2;
				read_binary_command(n);
				if (execute_binary_command(n)) return;
			}
		} else if (c == *p) {		// text command
				p++;
				if (*p == '\0') {
					read_text_command();
					if (execute_text_command()) return;
					p = pancake_header;
				}
		} else {
			p = pancake_header;
		}
	}

	/* NOTREACHED */
}
