// CC BY
// code by SASANO Takayoshi

#include "gfx.h"
#include "misc.h"

static	unsigned char	PixBuf[GFX_Y_SIZE][GFX_X_SIZE / 2];

unsigned char	gfx_get_pixel(signed char x, signed char y)
{
	unsigned char	p;

	if (x < 0 || x >= GFX_X_SIZE || y < 0 || y >= GFX_Y_SIZE) {
		return 0;
	}

	p = PixBuf[y][x / 2];

	if (x & 1) {
		return (p >> 0) & 0x0f;
	} else {
		return (p >> 4) & 0x0f;
	}
}

void	gfx_set_pixel(signed char x, signed char y, unsigned char color)
{
	unsigned char	c, p;

	if (x < 0 || x >= GFX_X_SIZE || y < 0 || y >= GFX_Y_SIZE) {
		return;
	}

	p = PixBuf[y][x / 2];

	if (x & 1) {
		c = (color & 0x0f) << 0;
		p &= 0xf0;
	} else {
		c = (color & 0x0f) << 4;
		p &= 0x0f;
	}

	PixBuf[y][x / 2] = c | p;
}

void	gfx_clear(unsigned char color)
{
	unsigned char	c;

	c = (color & 0x0f) | ((color & 0x0f) << 4);
	memset(PixBuf, c, sizeof(PixBuf));
}

void	gfx_draw_line(signed char x1, signed char y1, signed char x2, signed char y2, unsigned char color)
{
	long	x, y, dx, dy, n;

	if (x2 < x1) {
		n = x2;
		x2 = x1;
		x1 = n;
	}
	if (y2 < y1) {
		n = y2;
		y2 = y1;
		y1 = n;
	}

	dx = x2 - x1;
	dy = y2 - y1;
	n = x2 * y1 - x1 * y2;

	if (dx == 0 && dy == 0) {
		/* simply put pixel */
		gfx_set_pixel(x1, y1, color);
	} else if (dx > dy) {
		/* calculate y from x */
		for (x = x1; x <= x2; x++) {
			y = ((dy * x + n) * 2 + dx) / (dx * 2);
			gfx_set_pixel(x, y, color);
		}
	} else {
		/* calculate x from y */
		for (y = y1; y <= y2; y++) {
			x = ((dx * y - n) * 2 + dy) / (dy * 2);
			gfx_set_pixel(x, y, color);
		}
	}
}

void	gfx_draw_circle(signed char xc, signed char yc, signed char ra, unsigned char color)
{
	// reference: http://dencha.ojaru.jp/programs_07/pg_graphic_09a2.html

	long	x, y, d0, d1, xmax;

	/* xmax = ra * (sqrt(2) / 2) */
	xmax = (ra * 46341 + 32768) / 65536;

	for (x = 0, y = ra; x <= xmax; x++) {
		d0 = (x * x + y * y) - (ra * ra);
		d1 = (x * x + (y - 1) * (y - 1)) - (ra * ra);
		if (d0 < 0) d0 = -d0;
		if (d1 < 0) d1 = -d1;
		if (d0 > d1) y--;

		gfx_set_pixel(xc + x, yc + y, color);
		gfx_set_pixel(xc + x, yc - y, color);
		gfx_set_pixel(xc - x, yc + y, color);
		gfx_set_pixel(xc - x, yc - y, color);
		gfx_set_pixel(xc + y, yc + x, color);
		gfx_set_pixel(xc + y, yc - x, color);
		gfx_set_pixel(xc - y, yc + x, color);
		gfx_set_pixel(xc - y, yc - x, color);
	}
}

void	gfx_draw_block8x8(signed char x, signed char y, unsigned char key, const unsigned char *pattern)
{
	signed char 	px, py;
	unsigned char	c;

	if (pattern == NULL) {
		return;
	}

	for (py = 0; py < 8; py++) {
		for (px = 0; px < 8; px += 2) {
			c = (*pattern >> 4) & 0x0f;
			if (c != key) {
				gfx_set_pixel(x + px + 0, y + py, c);
			}

			c = (*pattern >> 0) & 0x0f;
			if (c != key) {
				gfx_set_pixel(x + px + 1, y + py, c);
			}

			pattern++;
		}
	}
}

void	gfx_draw_bitmap8x8(signed char x, signed char y, unsigned char color, const unsigned char *pattern)
{
	signed char	px, py;
	unsigned char	mask;

	if (pattern == NULL) {
		return;
	}

	for (py = 0; py < 8; py++) {
		mask = 0x80;
		for (px = 0; px < 8; px++) {
			if (*pattern & mask) {
				gfx_set_pixel(x + px, y + py, color);
			}
		}

		pattern++;
	}
}

void	gfx_display_image(const unsigned char *image)
{
	if (image == NULL) {
		return;
	}

	memcpy(PixBuf, image, sizeof(PixBuf));
}
