// CC BY
// code by SASANO Takayoshi

#include "io.h"
#include "intc.h"
#include "intentry.h"
#include "debug.h"

#define	TXBUF_SIZE	32
#define	RXBUF_SIZE	224
#define	TXFIFO_SIZE	16
#define	RXFIFO_SIZE	16

#define	COM_PORT	0x3f8			// COM1
#define	COM_IRQ		4
#define	COM_CLOCK	1843200			// Hz
#define	COM_SPEED	115200
#define	COM_DIVISOR	(COM_CLOCK / (COM_SPEED * 16))

#define	RHR		(COM_PORT + 0x00)	// r
#define	THR		(COM_PORT + 0x00)	//  w
#define	IER		(COM_PORT + 0x01)	// rw
#define	ISR		(COM_PORT + 0x02)	// r
#define	FCR		(COM_PORT + 0x02)	//  w
#define	LCR		(COM_PORT + 0x03)	// rw
#define	MCR		(COM_PORT + 0x04)	// rw
#define	LSR		(COM_PORT + 0x05)	// r
#define	MSR		(COM_PORT + 0x06)	// r

#define	DLL		(COM_PORT + 0x00)	// rw
#define	DLM		(COM_PORT + 0x01)	// rw

static	unsigned char	TxBuf[TXBUF_SIZE];
static	unsigned char	RxBuf[RXBUF_SIZE];

volatile static	unsigned char	TxStart;
volatile static	unsigned char	TxEnd;
volatile static	unsigned char	TxBusy;
volatile static	unsigned char	RxStart;
volatile static	unsigned char	RxEnd;

static	unsigned long	OldVector;

#define	TxPtr(x)	((x) % TXBUF_SIZE)
#define	RxPtr(x)	((x) % RXBUF_SIZE)

static	short	sio_write_buffer(unsigned char c)
{
	short	result;

	DI();
	if (TxPtr(TxEnd + 1) != TxStart) {
		if (TxEnd == TxStart) {
			out_b(IER, in_b(IER) | 0x02);	// enable tx interrupt
			TxBusy = 1;
		}

		TxBuf[TxEnd] = c;
		TxEnd = TxPtr(TxEnd + 1);

		result = 0;
	} else {
		/* no more buffer */
		result = -1;
	}
	EI();

	return result;
}

static	short	sio_read_buffer(void)
{
	short	result;

	DI();
	if (RxStart != RxEnd) {
		result = RxBuf[RxStart];
		RxStart = RxPtr(RxStart + 1);
	} else {
		/* no data received */
		result = -1;
	}
	EI();

	return result;
}

void	sio_interrupt()
{
	unsigned char	i, s, c;

	/* receive */
	for (i = 0; i < RXFIFO_SIZE; i++) {
		s = in_b(LSR);			// get status
		if (!(s & 0x01)) break;		// no data

		c = in_b(RHR);			// read from FIFO
		if (s & 0x1e) continue;		// error, discard

		if (RxPtr(RxEnd + 1) != RxStart) {
			RxBuf[RxEnd] = c;
			RxEnd = RxPtr(RxEnd + 1);
		}
	}

	/* transmit */
	if (TxBusy) {
		s = in_b(LSR);
		if (s & 0x20) {	// tx ready
			if (TxStart != TxEnd) {
				/* fill tx FIFO */
				for (i = 0; i < TXFIFO_SIZE; i++) {
					out_b(THR, TxBuf[TxStart]);
					TxStart = TxPtr(TxStart + 1);
					if (TxStart == TxEnd) break;
				}
			} else {
				/* no data to send, disable tx interrupt */
				out_b(IER, in_b(IER) & ~0x02);
				TxBusy = 0;
			}
		}
	}

	/* clear modem status interrupt */
	(void)in_b(MSR);

	send_eoi(COM_IRQ);
}

void	sio_send_char(unsigned char c)
{
	while (sio_write_buffer(c) < 0) {
		halt();
	}
}

unsigned char	sio_recv_char(void)
{
	short	c;

	while ((c = sio_read_buffer()) < 0) {
		halt();
	}
	debug_put_hex(c);	// XXX

	return c;
}

void	sio_init(unsigned char start)
{
	DI();

	disable_irq(COM_IRQ);

	out_b(IER, 0x00);		// disable interrupt
	while (!(in_b(LSR) & 0x40));	// wait for tx complete
	out_b(LCR, 0x80);		// set divisor
	out_b(DLM, 0xff);		// (avoid divisor = 0)
	out_b(DLL, (COM_DIVISOR >> 0) & 0xff);
	out_b(DLM, (COM_DIVISOR >> 8) & 0xff);
	out_b(LCR, 0x03);		// data 8bit, stop 1bit, non-parity
	out_b(MCR, 0x0b);		// RTS#, DTR# assert, INTR enable
	out_b(FCR, 0x87);		// FIFO enable and reset

	TxStart = TxEnd = TxBusy = 0;
	RxStart = RxEnd = 0;

	if (start) {
		OldVector = get_irq_vector(COM_IRQ);	// save old vector
		set_irq_vector(COM_IRQ,
			       (get_cs() << 16) | 
			       ((unsigned long)sio_interrupt_entry & 0xffff));
		out_b(IER, 0x05);	// enable rx ready, line status
		enable_irq(COM_IRQ);
	} else {
		set_irq_vector(COM_IRQ, OldVector);	// restore old vector
	}

	EI();
}
